/****************************************************************** 
 * Copyright (c) 2003, Mario Hochreiter, Student of 
 * Software Engineering at FH-Hagenberg, Austria. 
 * Project member of the isac team at the Institute for 
 * Software-Technologie, Graz University of Technology, Austria. 
 * * 
 * Use is subject to license terms.
 ******************************************************************/
package isac.gui;

import isac.bridge.CalcTree;
import isac.gui.treetable.CalculationModel;
import isac.gui.treetable.CustomTreeCellRenderer;
import isac.gui.util.CalcModelHierarchy;
import isac.gui.util.CalcModelNode;
import isac.util.CalcChangedEvent;
import isac.util.Formalization;
import isac.util.formulae.CalcHead;
import isac.util.formulae.CalcHeadCompoundID;
import isac.util.formulae.CalcFormula;
import isac.util.interfaces.ICalcIterator;
import isac.util.interfaces.IToUser;
import isac.util.interfaces.IToWorksheet;
import isac.util.tactics.Tactic;
import isac.wsdialog.DGuide;
import isac.wsdialog.DialogProtocolException;
import isac.wsdialog.IUIElement;
import isac.wsdialog.IUserAction;
import isac.wsdialog.UserAction;
import isac.wsdialog.UserActionOnCalcElement;
import isac.wsdialog.UserActionOnCalcHeadCompoundID;
import isac.wsdialog.UserActionOnIterator;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Locale;

import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.event.CellEditorListener;
import javax.swing.event.ChangeEvent;
import javax.swing.tree.DefaultTreeCellEditor;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeSelectionModel;

import org.apache.log4j.Logger;

/**
 * <code>Worheet</code> is the component that is responsible for the
 * interaction with the user while doing a calculation. Furthermore it acts as a
 * visual represantiation of mathematical formulas in a calculation tree, that
 * is beeing executed by the isac knowledge interpreter.
 * 
 * The <code>Worksheet</code> is the view of the
 * <code>WorksheetDialogGuide</code>.
 * 
 * The data, that the <code>WorksheetDialogGuide</code> delivers, is
 * representated in the hierarchical structur <code>JTree</code>.
 * 
 * @author Mario Hochreiter
 * @version 0.1 15/09/2003
 */
public class Worksheet extends JPanel implements ActionListener, /*IToUser, */ CellEditorListener, Serializable{

	private MouseAdapter ml;

	private JMenuItem showTactics;

	private JPopupMenu popupMenu;

	static Logger logger = Logger.getLogger(Worksheet.class.getName());

  protected IToUser worksheet_rmi_1_;
  protected IToWorksheet worksheet_rmi_2_;
  
  /** variable that says if the calculation has ended */
  protected boolean calculationEnded;
  
  // formula which will be edited if the dialog allows it
  protected CalcFormula edited_formula_;

  /** Model for the JTree */
  protected TreeModel model;
  
  protected CalcTree calcTree;

  protected CalcModelHierarchy calc;

  /** Used to represent the model. */
  protected JTree tree;

  /** Dialog which is responsible for the communication with the kernel */
  protected DGuide dialog;

  /** Button for doing the next step in calculation */
  //protected JButton nextStep;
  /** Calculate whole example (without single stepping) */
  //protected JButton autoCalc;
  /** true if it is allowed to call tryMatch/tryRefine, false otherwiese */
  protected boolean callProblemAllowed;

  /** true if in single stepping mode */
  protected boolean singleStepping;

  protected JPanel buttonPanel;

  /** Panel which is shows the calcHead if forced */
  protected CalcHeadPanel chp;

  /** Panle which shows the tactic if forced */
  protected TacticPanel tp;

  /** submit calcHead and start calculation */
  protected JButton startSolving;

  /** check calc head */
  protected JButton checkCalcHead;

  /** complete calc head */
  protected JButton compCalcHead;

  protected ICalcIterator it_;

  protected CalcHead ch;
  
  //LK 04.11.18
  protected int current_row_;

  /**
   * Create a <code>Worksheet</code> that displays the Formalization in the
   * requested viewStyle. The text of the labels is localized. The given
   * <code>DGuide</code> is connected to the <code>Worksheet</code>.
   * 
   * @param l
   *          Locale that is used.
   * @param dialogGuide
   *          The dialog guide that is connected to the Worksheet
   * @param f
   *          Formalization necessary to start a new calculation
   * @param source
   *          Where was the calculation started
   * @param viewStyle
   *          Requested view style (how detailed the CalcHead should be
   *          displayed)
   */
  public Worksheet(final Locale l, DGuide dialogGuide, Formalization f,
                   int source, int viewStyle) {
    // LK 04.11.18 (1 line)
    current_row_ = 0;
    singleStepping = true;
    callProblemAllowed = true;
    super.setLayout(new BorderLayout());
    dialog = dialogGuide;
    try {
      logger.debug(f.toString());
      calcTree = dialog.startCalculation(0, f, source, viewStyle);
      ICalcIterator it = calcTree.iterator();
      try {
        ch =  (CalcHead) it.getFormula();
        //dialog.setNextTactic(dialog.fetchProposedTactic());
      } catch (Exception e) {
        e.printStackTrace();
      }
      logger.debug(ch);
      logger.info("reqViewStyle: " + viewStyle);
      //TODO: alter
      // that's how the viewStyle should be set but it does not work
      //dialog.notifyUserAction(new
      // UserActionOnInt(IUserAction.UI_SPECIFY_CHANGE_VIEW, 5));
      //workaround
      ch.setViewStyle(CalcHead.VIEWSTYLE_FULL);
      logger.info("getViewStyle: " + ch.getViewStyle());
      logger.debug(ch);
      worksheet_rmi_1_ = new WorksheetRMI(this);
      worksheet_rmi_2_ = new WorksheetRMI(this);
      dialog.addDataChangeListener(worksheet_rmi_1_);
      dialog.registerUIControlListener(worksheet_rmi_2_);
      chp = new CalcHeadPanel(l, ch, ch.getViewStyle());
      this.buildCalcHeadPanelButtons();
      super.add(chp, BorderLayout.NORTH);
      this.startSolving();
    } catch (RemoteException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
      //    } catch (DialogProtocolException e) {
      //		logger.error(e);
    }
    //    outputText = new JTextArea(10, 50);
    //    this.add(new JScrollPane(outputText), BorderLayout.CENTER);
  }

  private void buildCalcHeadPanelButtons() {
    buttonPanel = new JPanel();
    //TODO localize!!!
    startSolving = new JButton("start Solving");
    compCalcHead = new JButton("complete CalcHead");
    checkCalcHead = new JButton("check CalcHead");
    startSolving.addActionListener(this);
    compCalcHead.addActionListener(this);
    checkCalcHead.addActionListener(this);
    buttonPanel.add(checkCalcHead);
    buttonPanel.add(compCalcHead);
    buttonPanel.add(startSolving);
    super.add(buttonPanel, BorderLayout.SOUTH);
  }

  //  private void buildButtons() {
  //    this.remove(buttonPanel);
  //    buttonPanel = new JPanel();
  //    //TODO localize!!!
  //    nextStep = new JButton("Next step");
  //	  autoCalc = new JButton("Autocalc");
  //    nextStep.addActionListener(this);
  //  	autoCalc.addActionListener(this);
  //    buttonPanel.add(nextStep);
  //    buttonPanel.add(autoCalc);
  //    this.add(buttonPanel, BorderLayout.SOUTH);
  //  }

  protected void buildMenu() {
    popupMenu = new JPopupMenu();
    showTactics = new JMenuItem("show applicable tactics");
    showTactics.addActionListener(this);
    popupMenu.add(showTactics);
    this.initMouseListener();
  }

  protected void initMouseListener() {
    ml = new MouseAdapter() {
      public void mousePressed(MouseEvent e) {
        logger.info("mouseEvent");
        if (e.getButton() == MouseEvent.BUTTON3 && !isCallProblemAllowed()) {
          popupMenu.show(e.getComponent(), e.getX(), e.getY());
        }
      }
    };
    tree.addMouseListener(ml);
  }

  private void showTacticPanel() {
    if (tp == null) {
      tp = new TacticPanel();
      super.add(tp, BorderLayout.EAST);
    }
  }

  /* LK 04.09.02 */
  public boolean isCalculationEnded() {
    return this.calculationEnded;
  }

  //TreeView
  public void setTreeTableModel() {
    calc = new CalcModelHierarchy();
    model = new CalculationModel(calc, calcTree.iterator());
    tree = new JTree(model);

    tree.setEditable(true);
    tree.setCellRenderer(new CustomTreeCellRenderer());
    tree.setCellEditor(new DefaultTreeCellEditor(tree,
        new DefaultTreeCellRenderer()));
    //LK 04.11.17: (1 line)
    tree.getCellEditor().addCellEditorListener(this);
    tree.getSelectionModel()
        .setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
    tree.setRootVisible(false);
    tree.setShowsRootHandles(true);
    KeyStroke ks = KeyStroke.getKeyStroke(KeyEvent.VK_A,
                                          KeyEvent.CTRL_DOWN_MASK);
    tree.getInputMap().put(ks, "none");
    super.add(new JScrollPane(tree), BorderLayout.CENTER);
  }

  /**
   * Do matching (see UC 2.5.6).
   * 
   * @param problemID
   *          unique id for a CalcHead
   */
  public void tryMatch(CalcHeadCompoundID problemID) {
    try {
      // dialog.tryMatch(ch, problemID);
      dialog.notifyUserAction(new UserActionOnCalcHeadCompoundID(
          IUIElement.UI_SPECIFY_TRY_MATCH, problemID));
      logger.debug("CalcHead after tryMatch: " + ch.toSMLString());
      chp.fillTextFieldsFromCalcHead(ch);
      this.showCalcHeadPane();
    } catch (RemoteException e) {
      e.printStackTrace();
      //    } catch (NotInSpecificationPhaseException e) {
      //      e.printStackTrace();
    } catch (DialogProtocolException e) {
      logger.error(e.getMessage(), e);
    }
  }

  /**
   * Refine a problem (see UC 2.5.7).
   * 
   * @param problemID
   *          unique id for a CalcHead
   */
  // method at the moment only gets CalcHead but does not
  // mark the refined problem in the problem hierarchy
  public void tryRefine(CalcHeadCompoundID problemID) {
    try {
      // dialog.tryRefine(ch, problemID);
      dialog.notifyUserAction(new UserActionOnCalcHeadCompoundID(
          IUIElement.UI_SPECIFY_TRY_REFINE, problemID));
      chp.fillTextFieldsFromCalcHead(ch);
      this.showCalcHeadPane();
    } catch (DialogProtocolException e) {
      logger.error(e);
    } catch (RemoteException e) {
      logger.error(e);
    }
  }

  /**
   * Show all applicable Tactics that can be used for the next calculation step
   * (UC 2.5.11).
   * 
   * @param scope
   *          The filter parameter selects the scope of search for applicable
   *          tactics.
   */
  public void showApplicableTactics(int scope) {
    try {
      Tactic[] tactic = dialog.fetchApplicableTactics(scope);
      //TODO: show tactis in a list the user can select
    } catch (RemoteException e) {
      logger.error(e);
    }
  }

  /**
   * Call problem is allowed if the calculation is not in solving phase.
   */
  public boolean isCallProblemAllowed() {
    return callProblemAllowed;
  }

  /**
   * @return true if the calculation has started.
   */
  public boolean isCalculationStarted() {
    return it_ != null;
  }

  private void setIterator() {
    try {
      it_ = calcTree.iterator();
      it_.moveRoot();
    } catch (RemoteException e) {
      e.printStackTrace();
      System.exit(5);
    }
  } 
  
  private void showCalcHeadPane() {
    super.add(chp, BorderLayout.NORTH);
    super.validate();
  }

  /* LK 04.09.01 */
  // to be called from the windowApplication
  public void nextStep() {
      singleStepping = true;
      Tactic propTactic;
      try {
        if (it_ == null)
          this.setIterator();
        propTactic = dialog.fetchProposedTactic();
        //FIXME:WN: vvvv ...endCalculation
        if (propTactic != null) {
          dialog.setNextTactic(propTactic);
          //((CalculationModel) model).nextStep();
        } else {
          this.calculationEnded = true;
          // nextStep.setEnabled(false);
          // autoCalc.setEnabled(false);
          // ((CalculationModel) model).nextStep();
        }
      } catch (RemoteException e1) {
        e1.printStackTrace();
      } //try
  }

  /* LK 04.09.01 */
  // to be called from the windowApplication
  public void autoCalc() {
    if (!this.calculationEnded) {
      try {
        singleStepping = false;
        dialog
              .notifyUserAction(new UserAction(
                                               IUserAction.UI_SOLVE_CALCULATE_ALL));
//        ((CalculationModel) model).nSteps();
        for (int i=0; i < tree.getRowCount(); i++) {
          tree.expandRow(i);
        }
        this.calculationEnded = true;
      } catch (RemoteException e1) {
        logger.error("actionPerformed", e1.detail);
      } catch (DialogProtocolException e1) {
        logger.error("actionPerformed");
      } //try
    }//if
  }

  public void startSolving() {
    try {
      //    dialog.newCalculation(ch);
      chp.fillCalcHeadFromTextFields(ch);
      if (dialog.notifyUserAction(new UserAction(
          IUserAction.UI_SPECIFY_CALCULATE_1))) {
        this.setIterator();
        this.setTreeTableModel();
        super.remove(buttonPanel);
        this.buildMenu();
        super.remove(chp);
        //		this.showTacticPanel();
        this.validate();
        callProblemAllowed = false;
      } else {
        JOptionPane.showMessageDialog(this, "Please enter a correct CalcHead",
                                      "CalcHead not correct",
                                      JOptionPane.WARNING_MESSAGE);
      }
    } catch (RemoteException e1) {
      logger.error("actionPerformed", e1.detail);
    } catch (DialogProtocolException e2) {
      logger.error("actionPerformed", e2);
    }
  }

  public void compCalcHead() {
    try {
      chp.fillCalcHeadFromTextFields(ch);
      dialog.notifyUserAction(new UserAction(
          IUserAction.UI_SPECIFY_COMPLETE_CALCHEAD));
      chp.fillTextFieldsFromCalcHead(ch);
    } catch (Exception e1) {
      logger.error(e1);
    }
  }

  public void checkCalcHead() {
    try {
      logger.debug(dialog);
      dialog.notifyUserAction(new UserAction(
          IUserAction.UI_SPECIFY_CHECK_CALCHEAD));
      chp.fillTextFieldsFromCalcHead(ch);
    } catch (Exception e1) {
      logger.error(e1);
    }
  }

  public void showTactics() {
    try {
      dialog.notifyUserAction(new UserAction(
          IUserAction.UI_SOLVE_GET_APPLICABLE_TACTICS));
    } catch (Exception e1) {
      logger.error(e1);
    }
  }
  
  //------------- Implementation of ActionListener ----------------
  public void actionPerformed(ActionEvent e) {
    if (e.getSource() == startSolving) {
      startSolving();
    } else if (e.getSource() == compCalcHead) {
      compCalcHead();
    } else if (e.getSource() == checkCalcHead) {
      checkCalcHead();
    } else if (e.getSource() == showTactics) {
      showTactics();
    } else { //unexpected event source
      logger.error("unexpected event source " + e.getSource());
    }
  }

  public void calcChanged(CalcChangedEvent event){
    logger.debug("single Stepping: " + singleStepping);
    logger.debug(model);
    logger.debug("calc" + calc);
    logger.debug(model);
    ((CalculationModel) model).calcChanged(event);
    for (int i=0; i < tree.getRowCount(); i++) {
      tree.expandRow(i);
    }
    tree.setEditable(true);
  }

  /**
   * if the user has canceled the edition, the worksheet should inform the dialogGuide
   * @see javax.swing.event.CellEditorListener#editingCanceled(javax.swing.event.ChangeEvent)
   */
  public void editingCanceled(ChangeEvent e) {
    // TODO Auto-generated method stub
    
  }

  /**
   * if the user has changed the value of a node, it should be altered
   * @see javax.swing.event.CellEditorListener#editingStopped(javax.swing.event.ChangeEvent)
   */
  public void editingStopped(ChangeEvent e) {
    Object o = tree.getCellEditor().getCellEditorValue();
    //TODO: LK 04.11.17: temporary solution: 

//    if (!((CalculationModel)model).editNode(current_node, (String)tree.getCellEditor().getCellEditorValue())) {
//      JOptionPane.showMessageDialog(this, "unable to find a way to this step",
//                                      "Error!", JOptionPane.ERROR_MESSAGE);
//    }
      //tree.getModel().valueForPathChanged(tree.getSelectionPath()), o);
    //LK 04.11.17: end of temporary solution:
    try {
      CalcModelNode current_node = (CalcModelNode)tree.getSelectionPath().getLastPathComponent();
      it_ = current_node.getValue();
      if (it_ == null) { // append formula
        //appends after the currently active formula
        edited_formula_ = new CalcFormula();
        edited_formula_.setText((String)o);
        logger.fatal("WS->DG: UI_SOLVE_APPEND_USER_FORMULA");
        dialog.notifyUserAction(new UserAction(IUserAction.UI_SOLVE_APPEND_USER_FORMULA));
      } else { //replace formula
        logger.fatal("WS->DG: UI_SOLVE_MOVE_ACTIVE_FORMULA");
        dialog.notifyUserAction(new UserActionOnIterator(IUserAction.UI_SOLVE_MOVE_ACTIVE_FORMULA, it_));
        edited_formula_ = new CalcFormula();
        edited_formula_.setText((String)o);
        edited_formula_.setPosition(it_.getPosition());
        logger.fatal("WS->DG: UI_SOLVE_EDIT_ACTIVE_FORMULA");
        dialog.notifyUserAction(new UserAction(IUserAction.UI_SOLVE_EDIT_ACTIVE_FORMULA));
      }
    } catch (Exception e1) {
      e1.printStackTrace();
    }
    System.out.println(o);
  }

  /**
   * @param action
   */
  public boolean doUIAction(IUserAction action) {
    int actionID = action.getActionID();
    switch (actionID) {
      case IUIElement.UI_DO_EDIT_FORMULA:
        try {
          logger.fatal("WS->DG: UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE");
          dialog.notifyUserAction(new UserActionOnCalcElement(IUserAction.UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE, edited_formula_));
        } catch (RemoteException e) {
          e.printStackTrace();
        } catch (DialogProtocolException e) {
          e.printStackTrace();
        }
        return true;
      case IUIElement.UI_DO_APPEND_FORMULA:
        try {
          logger.fatal("WS->DG: UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE");
          dialog.notifyUserAction(new UserActionOnCalcElement(IUserAction.UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE, edited_formula_));
        } catch (RemoteException e) {
          e.printStackTrace();
        } catch (DialogProtocolException e) {
          e.printStackTrace();
        }
      default:
        return false;
    }
      
  }
  
}